"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import datetime
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules import RuleMatch


class Base(CloudFormationLintRule):
    """Check Base Template Settings"""
    id = 'E1001'
    shortdesc = 'Basic CloudFormation Template Configuration'
    description = 'Making sure the basic CloudFormation template components are properly configured'
    source_url = 'https://github.com/aws-cloudformation/cfn-python-lint'
    tags = ['base']

    required_keys = [
        'Resources'
    ]

    def __init__(self):
        """Init"""
        super(Base, self).__init__()
        self.config_definition = {
            'sections': {
                'default': '',
                'type': 'string'
            }
        }
        self.configure()

    def _validate_version(self, template):
        results = []
        valid_version = '2010-09-09'
        if 'AWSTemplateFormatVersion' in template:
            version = template.get('AWSTemplateFormatVersion')
            if not isinstance(version, (str, datetime.date)):
                message = 'AWSTemplateFormatVersion only valid value is {0}'
                results.append(RuleMatch(['AWSTemplateFormatVersion'], message.format(valid_version)))
            else:
                if version != valid_version and version != datetime.datetime.strptime(valid_version, '%Y-%m-%d').date():
                    message = 'AWSTemplateFormatVersion only valid value is {0}'
                    results.append(RuleMatch(['AWSTemplateFormatVersion'], message.format(valid_version)))
        return results

    def _validate_transform(self, transforms):
        results = []
        if not isinstance(transforms, (list, str)):
            message = 'Transform has to be a list or string'
            results.append(RuleMatch(['Transform'], message.format()))
            return results

        if not isinstance(transforms, list):
            transforms = [transforms]

        return results

    def match(self, cfn):
        matches = []

        top_level = []
        for x in cfn.template:
            top_level.append(x)
            if x not in cfn.sections and x != self.config['sections']:
                message = 'Top level template section {0} is not valid'
                matches.append(RuleMatch([x], message.format(x)))

        for y in self.required_keys:
            if y not in top_level:
                message = 'Missing top level template section {0}'
                matches.append(RuleMatch([y], message.format(y)))

        matches.extend(self._validate_version(cfn.template))
        if 'Transform' in cfn.template:
            matches.extend(self._validate_transform(cfn.template.get('Transform')))
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules import RuleMatch


class BackupPlanLifecycleRule(CloudFormationLintRule):
    """Check Backup Plan rules with lifecycle has minimum period between cold and delete"""
    id = 'E3504'
    shortdesc = 'Check minimum 90 period is met between BackupPlan cold and delete'
    description = 'Check that Backup plans with lifecycle rules have >= 90 days between cold and delete'
    source_url = 'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-backup-backupplan-lifecycleresourcetype.html'
    tags = ['properties', 'backup', 'plan', 'lifecycle']

    def match(self, cfn):
        """Check cold storage and deletion lifecycle period differences"""
        matches = []
        results = cfn.get_resource_properties(['AWS::Backup::BackupPlan', 'BackupPlan', 'BackupPlanRule', 'Lifecycle'])

        for result in results:
            backup_rule = result['Value']
            # if 'MoveToColdStorageAfterDays' in backup_rule and 'DeleteAfterDays' in backup_rule:
            if isinstance(backup_rule.get('MoveToColdStorageAfterDays'), int) and isinstance(backup_rule.get('DeleteAfterDays'), int):
                if backup_rule['DeleteAfterDays'] - backup_rule['MoveToColdStorageAfterDays'] < 90:
                    message = 'DeleteAfterDays in {0} must be at least 90 days after MoveToColdStorageAfterDays'
                    rule_match = RuleMatch(result['Path'], message.format('/'.join(map(str, result['Path']))))
                    matches.append(rule_match)

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import re
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules import RuleMatch


class PreviousGenerationInstanceType(CloudFormationLintRule):
    id = 'I3100'
    shortdesc = 'Checks for legacy instance type generations'
    description = 'New instance type generations increase performance and decrease cost'
    source_url = 'https://aws.amazon.com/ec2/previous-generation/'
    tags = ['resources', 'ec2', 'rds', 'elasticcache', 'elasticsearch']

    def match(self, cfn):
        matches = []
        for resource_type, property_type in [
          ('AWS::AutoScaling::LaunchConfiguration', 'InstanceType'),
          ('AWS::EC2::CapacityReservation', 'InstanceType'),
          ('AWS::EC2::Host', 'InstanceType'),
          ('AWS::EC2::Instance', 'InstanceType'),
          ('AWS::RDS::DBInstance', 'DBInstanceClass'),
          ('AWS::ElastiCache::CacheCluster', 'CacheNodeType'),
          ('AWS::ElastiCache::GlobalReplicationGroup', 'CacheNodeType'),
          ('AWS::ElastiCache::ReplicationGroup', 'CacheNodeType'),
        ]:
            for resource_name, resource in cfn.get_resources([resource_type]).items():
                if isinstance(resource.get('Properties', {}).get(property_type, ''), str):
                    if re.search(r'([cmr][1-3]|cc2|cg1|cr1|g2|hi1|hs1|i2|t1)\.', resource.get('Properties', {}).get(property_type, '')):
                        matches.append(RuleMatch(['Resources', resource_name, 'Properties', property_type], 'Upgrade previous generation instance type: ' + resource.get('Properties').get(property_type)))

        for resource_type, top_level_property_type, property_type in [
          ('AWS::EC2::EC2Fleet', 'FleetLaunchTemplateOverridesRequest', 'InstanceType'),
          ('AWS::EC2::LaunchTemplate', 'LaunchTemplateData', 'InstanceType'),
          ('AWS::EC2::SpotFleet', 'SpotFleetLaunchSpecification', 'InstanceType'),
          ('AWS::OpenSearchService::Domain', 'ClusterConfig', 'InstanceType'),
          ('AWS::Elasticsearch::Domain', 'ElasticsearchClusterConfig', 'InstanceType'),
        ]:
            for resource_name, resource in cfn.get_resources([resource_type]).items():
                if isinstance(resource.get('Properties', {}).get(top_level_property_type, {}).get(property_type, ''), str):
                    if re.search(r'([cmr][1-3]|cc2|cg1|cr1|g2|hi1|hs1|i2|t1)\.', resource.get('Properties', {}).get(top_level_property_type, {}).get(property_type, '')):
                        matches.append(RuleMatch(['Resources', resource_name, 'Properties', top_level_property_type, property_type], 'Upgrade previous generation instance type: ' + resource.get('Properties').get(top_level_property_type).get(property_type)))
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule
from cfnlint.languageExtensions import LanguageExtensions


class ToJsonString(CloudFormationLintRule):
    """Check if ToJsonString values are correct"""
    id = 'E1031'
    shortdesc = 'ToJsonString validation of parameters'
    description = 'Making sure Fn::ToJsonString is configured correctly'
    source_url = 'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference.html'
    tags = ['functions', 'toJsonString']

    def match(self, cfn):
        has_language_extensions_transform = cfn.has_language_extensions_transform()
        unsupported_pseudo_parameters = [
            'AWS::NotificationARNs'
        ]

        matches = []
        intrinsic_function = 'Fn::ToJsonString'
        fn_toJsonString_objects = cfn.search_deep_keys(intrinsic_function)

        for fn_toJsonString_object in fn_toJsonString_objects:
            tree = fn_toJsonString_object[:-1]
            fn_toJsonString_object_value = fn_toJsonString_object[-1]
            LanguageExtensions.validate_transform_is_declared(self, has_language_extensions_transform, matches, tree,
                                                              intrinsic_function)
            LanguageExtensions.validate_type(self, fn_toJsonString_object_value, matches, tree, intrinsic_function)
            LanguageExtensions.validate_pseudo_parameters(self, fn_toJsonString_object_value, matches, tree,
                                                          unsupported_pseudo_parameters, intrinsic_function)
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule
from cfnlint.languageExtensions import LanguageExtensions
from cfnlint.rules import RuleMatch


class Length(CloudFormationLintRule):
    """Check if Length values are correct"""
    id = 'E1030'
    shortdesc = 'Length validation of parameters'
    description = 'Making sure Fn::Length is configured correctly'
    source_url = 'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-length.html'
    tags = ['functions', 'length']

    def match(self, cfn):
        has_language_extensions_transform = cfn.has_language_extensions_transform()
        intrinsic_function = 'Fn::Length'
        matches = []

        fn_length_objects = cfn.search_deep_keys(intrinsic_function)
        for fn_length_object in fn_length_objects:
            tree = fn_length_object[:-1]
            LanguageExtensions.validate_transform_is_declared(self, has_language_extensions_transform, matches, tree, intrinsic_function)
            self.validate_type(fn_length_object, matches, tree)
            self.validate_ref(fn_length_object, matches, tree, cfn)
        return matches

    def validate_ref(self, fn_length_object, matches, tree, cfn):
        fn_length_value = fn_length_object[-1]
        if isinstance(fn_length_value, dict):
            if len(fn_length_value.keys()) != 1 or (list(fn_length_value.keys())[0] not in ['Ref', 'Fn::Split']):
                self.addMatch(matches, tree, 'Fn::Length expects either an array, a Ref to an array or Fn::Split, '
                                             'but found unexpected object under Fn::Length at {0}')
                return

            if 'Ref' in fn_length_value:
                if fn_length_value['Ref'] not in cfn.get_parameter_names():
                    self.addMatch(matches, tree, 'Fn::Length can only reference list parameters at {0}')
                else:
                    referenced_parameter = cfn.get_parameters().get(fn_length_value['Ref'])
                    parameter_type = referenced_parameter.get('Type')
                    if 'List' not in parameter_type:
                        self.addMatch(matches, tree, 'Fn::Length can only reference list parameters at {0}')

    def addMatch(self, matches, tree, message):
        matches.append(RuleMatch(tree[:], message.format('/'.join(map(str, tree)))))

    def validate_type(self, fn_length_object, matches, tree):
        fn_length_value = fn_length_object[-1]
        if not isinstance(fn_length_value, dict) and not isinstance(fn_length_value, list):
            message = 'Fn::Length needs a list or a reference to a list at {0}'
            matches.append(RuleMatch(tree[:], message.format('/'.join(map(str, tree)))))
